"""
High-level access to amati functionality.
"""

import importlib
import json
import sys
from pathlib import Path

import jsonpickle
from pydantic import BaseModel, ValidationError
from pydantic_core import ErrorDetails

# pylint: disable=wrong-import-position

sys.path.insert(0, str(Path(__file__).parent.parent))
from amati._resolve_forward_references import resolve_forward_references
from amati.file_handler import load_file

type JSONPrimitive = str | int | float | bool | None
type JSONArray = list["JSONValue"]
type JSONObject = dict[str, "JSONValue"]
type JSONValue = JSONPrimitive | JSONArray | JSONObject


def dispatch(data: JSONObject) -> tuple[BaseModel | None, list[ErrorDetails] | None]:
    """
    Returns the correct model for the passed spec

    Args:
        data: A dictionary representing an OpenAPI specification

    Returns:
        A pydantic model representing the API specification
    """

    version: JSONValue = data.get("openapi")

    if not isinstance(version, str):
        raise ValueError("A OpenAPI specification version must be a string.")

    if not version:
        raise ValueError("An OpenAPI Specfication must contain a version.")

    version_map: dict[str, str] = {
        "3.1.1": "311",
        "3.1.0": "311",
        "3.0.4": "304",
        "3.0.3": "304",
        "3.0.2": "304",
        "3.0.1": "304",
        "3.0.0": "304",
    }

    module = importlib.import_module(f"amati.validators.oas{version_map[version]}")

    resolve_forward_references(module)

    try:
        model = module.OpenAPIObject(**data)
    except ValidationError as e:
        return None, e.errors()

    return model, None


def check(original: JSONObject, validated: BaseModel) -> bool:
    """
    Runs a consistency check on the output of amati.
    Determines whether the validated model is the same as the
    originally provided API Specification

    Args:
        original: The dictionary representation of the original file
        validated: A Pydantic model representing the original file

    Returns:
        Whether original and validated are the same.
    """

    original_ = json.dumps(original, sort_keys=True)

    json_dump = validated.model_dump_json(exclude_unset=True, by_alias=True)
    new_ = json.dumps(json.loads(json_dump), sort_keys=True)

    return original_ == new_


def run(file_path: str, consistency_check: bool = False, store_errors: bool = False):
    """
    Runs the full amati process
    """

    data = load_file(file_path)

    result, errors = dispatch(data)

    if result and consistency_check:
        if check(data, result):
            print("Consistency check successful")
        else:
            print("Consistency check failed")

    if errors and store_errors:
        if not Path(".amati").exists():
            Path(".amati").mkdir()

        with open(".amati/pydantic.json", "w", encoding="utf-8") as f:
            f.write(jsonpickle.encode(errors, unpicklable=False))  # type: ignore


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(
        prog="amati",
        description="Test whether a OpenAPI specification is valid.",
    )

    parser.add_argument(
        "-s", "--spec", required=True, help="The specification to be parsed"
    )

    parser.add_argument(
        "-cc",
        "--consistency-check",
        required=False,
        action="store_true",
        help="Runs a consistency check between the input specification and amati",
    )

    parser.add_argument(
        "-se",
        "--store-errors",
        required=False,
        action="store_true",
        help="Stores and errors in a file for visibility.",
    )

    args = parser.parse_args()

    run(args.spec, args.consistency_check, args.store_errors)



"""
Tests amati/validators/oas311.py
"""

import json
import warnings
from pathlib import Path
from typing import Any

import pytest
import yaml

from amati.amati import check, dispatch, load_file
from amati.logging import LogMixin


def get_test_data() -> dict[str, Any]:
    """
    Gathers the set of test data.
    """

    with open("tests/data/.amati.tests.yaml", "r", encoding="utf-8") as f:
        content = yaml.safe_load(f)

    return content


def get_errors(error_file: Path) -> list[dict[str, Any]]:
    """
    Returns the stored, expected, set of errors associated
    with a given test specification.
    """

    with open(error_file, "r", encoding="utf-8") as f:
        expected_errors = json.loads(f.read())

    return expected_errors


@pytest.mark.external
def test_specs():

    content = get_test_data()

    directory = Path(content["directory"])

    for name, repo in content["repos"].items():

        for spec in repo["specs"]:
            file: Path = Path(directory) / name / spec

            data = load_file(file)

            with LogMixin.context():
                result, errors = dispatch(data)

            if errors := repo.get("error_file"):
                error_file = get_errors(Path(errors))

                assert json.dumps(error_file, sort_keys=True) == json.dumps(
                    error_file, sort_keys=True
                )
                assert not result

            else:

                assert not errors
                assert result

                # Pydantic emits a set of warnings with the error
                # PydanticSerializationUnexpectedValue when running this check. The
                # purpose of the check is to validate that amati can output the same
                # file that was provided. If the test passes then there has been no
                # incorrect serialisation.
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        action="ignore",
                        category=UserWarning,
                        message="Pydantic serializer warnings:.*",
                    )
                    assert check(data, result)



"""
Script to download the HTTP Authentication Schemes (ISO 9110) from IANA
"""

import csv
import io
import json
import pathlib

import requests

if __name__ == "__main__":
    response = requests.get(
        "https://www.iana.org/assignments/http-authschemes/authschemes.csv",
        timeout=20,
    )
    response.raise_for_status()  # Raise an exception for HTTP errors

    reader = csv.DictReader(io.StringIO(response.text))

    data: list[dict[str, str]] = []

    for row in reader:
        data.append(row)

    write_to = (
        pathlib.Path(__file__).parent.parent.resolve() / "amati/data/iso9110.json"
    )

    with write_to.open("w") as f:
        f.write(json.dumps(data))



"""
Validates an email according to the RFC5322 ABNF grammar - §3:
"""

from abnf import ParseError
from abnf.grammars import rfc5322

from amati import AmatiValueError, Reference
from amati.fields import Str as _Str

reference = Reference(
    title="Internet Message Format",
    url="https://www.rfc-editor.org/rfc/rfc5322#section-3",
    section="Syntax",
)


class Email(_Str):

    def __init__(self, value: str):

        try:
            rfc5322.Rule("address").parse_all(value)
        except ParseError as e:
            raise AmatiValueError(
                message=f"{value} is not a valid email address", reference=reference
            ) from e



"""
Amati is a specification validator, built to put a specification into
a single datatype and validate on instantiation.
"""

__version__ = "0.1.0"

# Imports are here for convenience, they're not going to be used here
# pylint: disable=unused-import
# pyright: reportUnusedImport=false

from amati.amati import dispatch, run
from amati.exceptions import AmatiValueError
from amati.references import AmatiReferenceException, Reference, References



"""
Tests amati/validators/oas311.py - ServerVariableObject
and the sub-objects OAuthFlowsObject
"""

import pytest
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.provisional import urls
from pydantic import ValidationError

from amati.fields import URI
from amati.fields.iso9110 import HTTP_AUTHENTICATION_SCHEMES
from amati.logging import LogMixin
from amati.validators.oas304 import OAuthFlowsObject
from amati.validators.oas311 import SecuritySchemeObject
from tests.helpers import text_excluding_empty_string

VALID_SECURITY_SCHEME_TYPES: list[str] = list(
    SecuritySchemeObject._SECURITY_SCHEME_TYPES  # pylint: disable=protected-access # type: ignore
)
INVALID_SECURITY_SCHEME_TYPES: st.SearchStrategy[str] = (
    st.text()
    .filter(lambda x: x not in VALID_SECURITY_SCHEME_TYPES)
    .filter(lambda x: x != "")
)

VALID_HTTP_AUTHENTICATION_SCHEMES: list[str] = list(HTTP_AUTHENTICATION_SCHEMES)
INVALID_HTTP_AUTHENTICATION_SCHEMES: st.SearchStrategy[str] = st.text().filter(
    lambda x: x not in VALID_HTTP_AUTHENTICATION_SCHEMES
)


@given(INVALID_SECURITY_SCHEME_TYPES)
def test_security_scheme_invalid(scheme_type: str):

    with LogMixin.context():
        SecuritySchemeObject(type=scheme_type)
        assert LogMixin.logs
        assert LogMixin.logs[0].message is not None
        assert LogMixin.logs[0].type == ValueError


@given(st.none())
def test_security_scheme_none(scheme_type: str):
    with pytest.raises(ValidationError):
        SecuritySchemeObject(type=scheme_type)


@given(
    st.text(),
    text_excluding_empty_string(),
    st.sampled_from(("query", "header", "cookie")),
)
def test_security_scheme_apikey_valid(description: str, name: str, in_: str):
    with LogMixin.context():
        SecuritySchemeObject(
            **{
                "type": "apiKey",
                "description": description,
                "name": name,
                "in": in_,
            }  # type: ignore
        )
        assert not LogMixin.logs


@given(
    st.text(),
    st.text(),
    st.text().filter(lambda x: x not in ("query", "header", "cookie")),
)
def test_security_scheme_apikey_invalid(description: str, name: str, in_: str):
    with LogMixin.context():
        SecuritySchemeObject(
            **{
                "type": "apiKey",
                "description": description,
                "name": name,
                "in": in_,
            }  # type: ignore
        )
        assert LogMixin.logs
        assert LogMixin.logs[0].message is not None
        assert LogMixin.logs[0].type == ValueError


@given(st.text(), st.sampled_from(VALID_HTTP_AUTHENTICATION_SCHEMES), st.text())
def test_security_scheme_http_valid(description: str, scheme: str, bearer_format: str):
    with LogMixin.context():
        SecuritySchemeObject(
            type="http",
            description=description,
            scheme=scheme,  # type: ignore
            bearerFormat=bearer_format,
        )
        assert not LogMixin.logs


@given(st.text(), st.text(), INVALID_HTTP_AUTHENTICATION_SCHEMES, st.text())
def test_security_scheme_http_invalid(
    type_: str, description: str, scheme: str, bearer_format: str
):
    with pytest.raises(ValidationError):
        SecuritySchemeObject(
            type=type_,
            description=description,
            scheme=scheme,  # type: ignore
            bearerFormat=bearer_format,
        )


@given(st.text(), st.sampled_from(VALID_HTTP_AUTHENTICATION_SCHEMES), st.text())
def test_security_scheme_oauth2_valid(
    description: str, scheme: str, bearer_format: str
):
    with LogMixin.context():
        SecuritySchemeObject(
            type="oauth2",
            description=description,
            scheme=scheme,  # type: ignore
            bearerFormat=bearer_format,
            flows=OAuthFlowsObject(),
        )
        assert not LogMixin.logs


@given(st.text(), st.text(), INVALID_HTTP_AUTHENTICATION_SCHEMES, st.text())
def test_security_scheme_oauth2_invalid(
    type_: str, description: str, scheme: str, bearer_format: str
):
    with pytest.raises(ValidationError):
        SecuritySchemeObject(
            type=type_,
            description=description,
            scheme=scheme,  # type: ignore
            bearerFormat=bearer_format,
        )


@given(
    urls(),
    urls(),
    st.dictionaries(keys=text_excluding_empty_string(), values=st.text()),
)
def test_oauth_flows_implicit_valid(
    authorization_url: URI, refresh_url: URI, scopes: dict[str, str]
):
    with LogMixin.context():
        OAuthFlowsObject(
            **{
                "implicit": {
                    "authorizationUrl": authorization_url,
                    "refreshUrl": refresh_url,
                    "scopes": scopes,
                }
            }  # type: ignore
        )

        assert not LogMixin.logs


@given(
    urls(),
    urls(),
    urls(),
    st.dictionaries(keys=text_excluding_empty_string(), values=st.text()),
)
def test_oauth_flows_implicit_invalid(
    authorization_url: URI, token_url: URI, refresh_url: URI, scopes: dict[str, str]
):
    with LogMixin.context():
        OAuthFlowsObject(
            **{
                "implicit": {
                    "authorizationUrl": authorization_url,
                    "tokenUrl": token_url,
                    "refreshUrl": refresh_url,
                    "scopes": scopes,
                }
            }  # type: ignore
        )

        assert LogMixin.logs


@given(
    urls(),
    urls(),
    urls(),
    st.dictionaries(keys=text_excluding_empty_string(), values=st.text()),
)
def test_oauth_flows_authorization_code_valid(
    authorization_url: URI, token_url: URI, refresh_url: URI, scopes: dict[str, str]
):
    with LogMixin.context():
        OAuthFlowsObject(
            **{
                "authorizationCode": {
                    "authorizationUrl": authorization_url,
                    "tokenUrl": token_url,
                    "refreshUrl": refresh_url,
                    "scopes": scopes,
                }
            }  # type: ignore
        )
        assert not LogMixin.logs


@given(
    urls(),
    urls(),
    st.dictionaries(keys=text_excluding_empty_string(), values=st.text()),
)
def test_oauth_flows_authorization_code_invalid(
    uri: URI, refresh_url: URI, scopes: dict[str, str]
):
    with LogMixin.context():
        OAuthFlowsObject(
            **{
                "authorizationCode": {
                    "authorizationUrl": uri,
                    "refreshUrl": refresh_url,
                    "scopes": scopes,
                }
            }  # type: ignore
        )
        assert LogMixin.logs
    with LogMixin.context():
        OAuthFlowsObject(
            **{
                "authorizationCode": {
                    "tokenUrl": uri,
                    "refreshUrl": refresh_url,
                    "scopes": scopes,
                }
            }  # type: ignore
        )
        assert LogMixin.logs


@given(
    urls(),
    urls(),
    st.dictionaries(keys=text_excluding_empty_string(), values=st.text()),
)
def test_oauth_flows_client_and_password_valid(
    token_url: URI, refresh_url: URI, scopes: dict[str, str]
):
    with LogMixin.context():
        OAuthFlowsObject(
            **{
                "clientCredentials": {
                    "tokenUrl": token_url,
                    "refreshUrl": refresh_url,
                    "scopes": scopes,
                }
            }  # type: ignore
        )

        assert not LogMixin.logs
    with LogMixin.context():
        OAuthFlowsObject(
            **{
                "password": {
                    "tokenUrl": token_url,
                    "refreshUrl": refresh_url,
                    "scopes": scopes,
                }
            }  # type: ignore
        )
        assert not LogMixin.logs


@given(
    urls(),
    urls(),
    urls(),
    st.dictionaries(keys=text_excluding_empty_string(), values=st.text()),
)
def test_oauth_flows_client_and_password_invalid(
    authorization_url: URI, token_url: URI, refresh_url: URI, scopes: dict[str, str]
):
    with LogMixin.context():
        OAuthFlowsObject(
            **{
                "clientCredentials": {
                    "authorizationUrl": authorization_url,
                    "tokenUrl": token_url,
                    "refreshUrl": refresh_url,
                    "scopes": scopes,
                }
            }  # type: ignore
        )
        assert LogMixin.logs
    with LogMixin.context():
        OAuthFlowsObject(
            **{
                "password": {
                    "authorizationUrl": authorization_url,
                    "tokenUrl": token_url,
                    "refreshUrl": refresh_url,
                    "scopes": scopes,
                }
            }  # type: ignore
        )
        assert LogMixin.logs



"""
Tests amati/references.py
"""

import pytest
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.provisional import urls

from amati import AmatiReferenceException, Reference


@given(st.text(), st.text(), urls())
def test_valid_reference_object(title: str, section: str, url: str):
    ref = Reference(title, section, url)

    assert ref.title == title
    assert ref.section == section
    assert ref.url == url


def test_invalid_reference_object():

    with pytest.raises(AmatiReferenceException):
        Reference()



"""
Tests amati/logging.py
"""

from amati import Reference, References
from amati.logging import Log, LogMixin
from amati.validators.generic import GenericObject

reference1 = Reference(title="Test", url="https://example.com")

reference2 = Reference(title="Test", url="https://a.com")

references: References = [reference1, reference2]


class Model1(GenericObject):
    value: str

    def test_log(self):
        LogMixin.log(Log(message="Model1", type=ValueError, reference=reference1))


class Model2(GenericObject):
    value: str

    def test_log(self):
        LogMixin.log(Log(message="Model2", type=ValueError, reference=references))


def test_writer():
    with LogMixin.context():
        model1 = Model1(value="a")
        model1.test_log()
        assert LogMixin.logs == [
            Log(message="Model1", type=ValueError, reference=reference1)
        ]

        model2 = Model2(value="b")
        model2.test_log()
        assert LogMixin.logs == [
            Log(message="Model1", type=ValueError, reference=reference1),
            Log(message="Model2", type=ValueError, reference=references),
        ]



"""
A generic object to add extra functionality to pydantic.BaseModel.

Should be used as the base class for all classes in the project.
"""

import re
from typing import (
    Any,
    Callable,
    ClassVar,
    Optional,
    Pattern,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseModel, ConfigDict, PrivateAttr
from pydantic_core._pydantic_core import PydanticUndefined

from amati import Reference
from amati.logging import Log, LogMixin


class GenericObject(LogMixin, BaseModel):
    """
    A generic model to overwrite provide extra functionality
    to pydantic.BaseModel.
    """

    _reference: ClassVar[Reference] = PrivateAttr()
    _extra_field_pattern: Optional[Pattern[str]] = PrivateAttr()

    def __init__(self, **data: Any) -> None:

        super().__init__(**data)

        if self.model_config.get("extra") == "allow":
            return

        # If extra fields aren't allowed log those that aren't going to be added
        # to the model.
        for field in data:
            if (
                field not in self.model_dump().keys()
                and field not in self.get_field_aliases()
            ):
                message = f"{field} is not a valid field for {self.__repr_name__()}."
                self.log(
                    Log(
                        message=message,
                        type=ValueError,
                    )
                )

    def model_post_init(self, __context: Any) -> None:
        if not self.model_extra:
            return

        if self.__private_attributes__["_extra_field_pattern"] == PrivateAttr(
            PydanticUndefined
        ):
            return

        # Any extra fields are allowed
        if self._extra_field_pattern is None:
            return

        excess_fields: set[str] = set()

        pattern: Pattern[str] = re.compile(self._extra_field_pattern)
        excess_fields.update(
            key for key in self.model_extra.keys() if not pattern.match(key)
        )

        for field in excess_fields:
            message = f"{field} is not a valid field for {self.__repr_name__()}."
            LogMixin.log(
                Log(
                    message=message,
                    type=ValueError,
                )
            )

    def get_field_aliases(self) -> list[str]:
        """
        Gets a list of aliases for confirming whether extra
        fields are allowed.

        Returns:
            A list of field aliases for the class.
        """

        aliases: list[str] = []

        for field_info in self.__class__.model_fields.values():
            if field_info.alias:
                aliases.append(field_info.alias)

        return aliases


T = TypeVar("T", bound=GenericObject)


def allow_extra_fields(pattern: Optional[str] = None) -> Callable[[Type[T]], Type[T]]:
    """
    A decorator that modifies a Pydantic BaseModel to allow extra fields and optionally
    sets a pattern for those extra fields

    Args:
        pattern: Optional pattern string for extra fields. If not provided all extra
        fields will be allowed

    Returns:
        A decorator function that adds a ConfigDict allowing extra fields
        and the pattern those fields should follow to the class.
    """

    def decorator(cls: Type[T]) -> Type[T]:
        """
        A decorator function that adds a ConfigDict allowing extra fields.
        """
        namespace: dict[str, Union[ConfigDict, Optional[str]]] = {
            "model_config": ConfigDict(extra="allow"),
            "_extra_field_pattern": pattern,
        }
        # Create a new class with the updated configuration
        return cast(Type[T], type(cls.__name__, (cls,), namespace))

    return decorator



"""
Represents a reference, declared here to not put in __init__.
"""

from dataclasses import dataclass
from typing import Optional, Sequence


class AmatiReferenceException(Exception):
    message: str = "Cannot construct empty references"


@dataclass
class Reference:
    """
    Attributes:
        title : Title of the referenced content
        section : Section of the referenced content
        url : URL where the referenced content can be found
    """

    title: Optional[str] = None
    section: Optional[str] = None
    url: Optional[str] = None

    def __post_init__(self):

        if not self.title and not self.section and not self.url:
            raise AmatiReferenceException


type ReferenceArray = Sequence[Reference]
type References = Reference | ReferenceArray



"""
Validates the idenfitier and licences from the System Package Data
Exchange (SPDX) licence list.
"""

import json
import pathlib

from amati import AmatiValueError, Reference
from amati.fields import Str as _Str
from amati.fields.uri import URI

reference = Reference(
    title="SPDX License List",
    url="https://spdx.org/licenses/",
)


DATA_DIRECTORY = pathlib.Path(__file__).parent.parent.resolve() / "data"

with open(DATA_DIRECTORY / "spdx-licences.json", "r", encoding="utf-8") as f:
    data = json.loads(f.read())

# `seeAlso` is the list of URLs associated with each licence
VALID_LICENCES: dict[str, list[str]] = {
    licence["licenseId"]: licence["seeAlso"] for licence in data["licenses"]
}
VALID_URLS: list[str] = [url for urls in VALID_LICENCES.values() for url in urls]


class SPDXIdentifier(_Str):
    """
    A class representing a valid SPDX license identifier.

    This class validates that a string value is a registered SPDX license identifier
    according to the official SPDX license list. It inherits from _Str to maintain
    compatibility with string operations while adding SPDX-specific validation.

    SPDX identifiers are standardized short-form identifiers for open source licenses,
    such as "MIT", "Apache-2.0", or "GPL-3.0-only".

    Attributes:
        Inherits all attributes from _Str

    Example:
        >>> license_id = SPDXIdentifier("MIT")
        >>> str(license_id)
        'MIT'
        >>> SPDXIdentifier("InvalidLicense")
        Traceback (most recent call last):
        amati.AmatiValueError: message
    """

    def __init__(self, value: str):

        if value not in VALID_LICENCES:
            raise AmatiValueError(
                f"{value} is not a valid SPDX licence identifier", reference=reference
            )


class SPDXURL(URI):  # pylint: disable=invalid-name
    """
    A class representing a valid SPDX license URL.

    This class validates that a URI is associated with an SPDX license in the official
    SPDX license list. It inherits from URI to maintain compatibility with URI
    validation and operations while adding SPDX-specific validation.

    SPDX license URLs are the official reference URLs for licenses in the SPDX registry,
    typically pointing to the canonical text of the license.

    Attributes:
        Inherits all attributes from URI

    Example:
        >>> license_url = SPDXURL("https://www.apache.org/licenses/LICENSE-2.0")
        >>> license_url.scheme
        'https'
        >>> SPDXURL("https://example.com")
        Traceback (most recent call last):
        amati.AmatiValueError: message
    """

    def __init__(self, value: str):

        super().__init__(value)

        if value not in VALID_URLS:
            raise AmatiValueError(
                f"{value} is not associated with any identifier.", reference=reference
            )
